# -*- coding: utf-8 -*-
import configparser
from fuclib import ezfuc
import os


class MyConfigParser(configparser.ConfigParser):
    def __init__(self, defaults=None):
        configparser.ConfigParser.__init__(self, defaults=defaults)

    def optionxform(self, optionstr):
        return optionstr


def get_config():
    current_path = os.path.abspath('../..')
    if 'conf' not in os.listdir(current_path):
        current_path = os.path.abspath('..')

    config_file_path = os.path.join(os.path.abspath(current_path + os.path.sep), 'conf', 'config.ini')
    conf = MyConfigParser()
    conf.read(config_file_path, encoding='utf-8')
    items = list()
    for section in conf.sections():
        items += conf.items(section)
    return ezfuc.dict_to_object(dict(items))


conf = get_config()



"""
Scrapy_Rabbit - a web crawling and web scraping framework written for Scrapy
"""

import pkgutil
import sys
import warnings
from scrapy_rabbit.utils import *
from scrapy_rabbit.core import *

from twisted import version as _txv

# Declare top-level shortcuts
from scrapy.spiders import Spider
from scrapy.http import Request, FormRequest
from scrapy.selector import Selector
from scrapy.item import Item, Field
from scrapy_rabbit.conf import conf

__all__ = [
    '__version__', 'version_info', 'twisted_version', 'Spider',
    'Request', 'FormRequest', 'Selector', 'Item', 'Field',
]

# Scrapy and Twisted versions
__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()
version_info = tuple(int(v) if v.isdigit() else v for v in __version__.split('.'))
twisted_version = (_txv.major, _txv.minor, _txv.micro)

# Check minimum required Python version
if sys.version_info < (3, 6):
    print("Scrapy %s requires Python 3.6+" % __version__)
    sys.exit(1)

# Ignore noisy twisted deprecation warnings
warnings.filterwarnings('ignore', category=DeprecationWarning, module='twisted')

del pkgutil
del sys
del warnings



# -*- coding: utf-8 -*-
class Pipeline:
    def __init__(self):
        self.count = 0

    def open_spider(self, spider):
        pass

    def process_item(self, item, spider):
        self.count += 1

    def close_spider(self, spider):
        pass



# -*- coding: utf-8 -*-
from importlib import import_module
import datetime
import logging
import time
import os
import sys
from scrapy_rabbit.utils import SpiderLoader
from scrapy_rabbit.utils import get_project_settings
from scrapy_rabbit.settings import overridden_settings

settings_base = get_project_settings()
settings_copy = settings_base.copy()
settings = dict(overridden_settings(settings_copy))
# log_level = settings.get('LOG_LEVEL', 'ERROR')
#
# save_log_file = settings.get('SAVE_LOG', False)
#
# if save_log_file:
#     logging.basicConfig(level=getattr(logging, log_level),
#                         format='pid:%(process)d %(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
#                         datefmt='%Y-%m-%d %H:%M:%S',
#                         filename=settings.get('LOG_FILE', f'{time.strftime("%Y-%m-%d", time.localtime())}.log'),
#                         filemode='a'
#                         )
# else:
#     logging.basicConfig(level=getattr(logging, log_level),
#                         format='pid:%(process)d %(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
#                         )  # todo log使用需要修改


root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(root)

logging.basicConfig(level=getattr(logging, getattr(settings, 'LOG_LEVEL', 'INFO')),
                    format='pid:%(process)d %(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    # filename='debug.txt',
                    # filemode='a'
                    )


def run(cmd) -> None:
    """手动启动方法"""
    if isinstance(cmd, str):
        cmd = cmd.split(' ')
    while '' in cmd:
        cmd.remove('')
    base_path = 'spiders.'
    if len(cmd) != 4:
        logging.error("命令行格式错误")
        raise
    try:
        path = base_path + cmd[0]
        root_path = cmd[0]
        spider_name = cmd[1]
        way = cmd[2]
        async_num = int(cmd[3])
        queue_name = cmd[0] + '/' + spider_name
    except Exception:
        logging.error("命令行格式错误")
        raise



    start_time = datetime.datetime.now()
    logging.info(f"爬虫启动时间：{start_time}")

    sl = SpiderLoader(type("settings", (object,), dict(getlist=lambda x: [path], getbool=lambda x: False)))
    spider_module = sl.load(spider_name)
    sp = spider_module(path=path, queue_name=queue_name, way=way, async_num=async_num)

    get_settings = getattr(spider_module, 'custom_settings', None)

    # a = {
    #     'a': 123,
    #     'b': 3231,
    #     'asd': 321,
    #     'c': 4325
    # }
    #
    # print(sorted(a.items(), key=lambda item: item[1]))
    # todo 多个pipline按顺序执行

    if get_settings:
        try:
            item_pipelines = list(get_settings['ITEM_PIPELINES'].keys())
            if '.' in item_pipelines[0]:
                pipline_name = item_pipelines[0].split('.')[-1]
            else:
                pipline_name = item_pipelines[0]
        except KeyError:
            pipline_name = "%sPipeline" % root_path
    else:
        pipline_name = "%sPipeline" % root_path

    pipeline = import_module("pipelines")
    pipelineObj = getattr(pipeline, pipline_name, getattr(pipeline, 'Pipeline'))()

    pipelineObj.open_spider(sp)
    setattr(sp, "pipelineObj", pipelineObj)
    sp.main()
    pipelineObj.close_spider(sp)
    end_time = datetime.datetime.now()

    logging.info(f"爬虫启动时间：{start_time}")
    logging.info(f"爬虫结束时间：{end_time}")
    logging.info(f"爬虫运行时长：{end_time - start_time}")
    logging.info(f"Item数量：{pipelineObj.count}")



"""Helper functions for working with templates"""

import os
import re
import string


def render_templatefile(path, **kwargs):
    with open(path, 'rb') as fp:
        raw = fp.read().decode('utf8')

    content = string.Template(raw).substitute(**kwargs)

    render_path = path[:-len('.tmpl')] if path.endswith('.tmpl') else path
    with open(render_path, 'wb') as fp:
        fp.write(content.encode('utf8'))
    if path.endswith('.tmpl'):
        os.remove(path)


CAMELCASE_INVALID_CHARS = re.compile(r'[^a-zA-Z\d]')


def string_camelcase(string):
    """ Convert a word  to its CamelCase version and remove invalid chars

    >>> string_camelcase('lost-pound')
    'LostPound'

    >>> string_camelcase('missing_images')
    'MissingImages'

    """
    return CAMELCASE_INVALID_CHARS.sub('', string.title())



# -*- coding: utf-8 -*-
import fuclib
import os
import scrapy
from scrapy_rabbit.conf import conf
from scrapy.utils.reqser import request_to_dict
from collections import Generator
from asyncio.events import new_event_loop
from asyncio.tasks import run_coroutine_threadsafe
from logging import getLogger
from threading import Thread
from scrapy_rabbit.http import Request
from scrapy_rabbit.utils import get_project_settings
from scrapy_rabbit.rabbitmq import RabbitConnect
from re import search
from pickle import dumps, loads
from pika import exceptions
from asyncio import set_event_loop
from time import ctime
from functools import partial
from traceback import print_exc
from scrapy_rabbit.http import random_ua
from scrapy_rabbit.settings import overridden_settings

logger = getLogger(__name__)


class DaemonRun(Thread):
    def __init__(self, func, loop, scrapy_object, *args, **kwargs):
        super(DaemonRun, self).__init__(*args, **kwargs)
        self.connector, self.channel = RabbitConnect(scrapy_object=scrapy_object).connect()
        self.func = func
        self.loop = loop
        self.setDaemon(True)

    def run(self) -> None:
        self.func(self.loop)


class Engine(object):
    import sys
    import os
    sys.path.append(os.path.abspath('../..'))
    sys.path.append(os.path.abspath('../../..'))


    timeout = 180
    max_times = 3
    start_urls = list()
    custom_settings = dict()
    allow_status_code = list()

    priority_queue = False
    queue_purge = True

    proxy = False

    def __init__(self, path, queue_name, way='auto', async_num=10):
        settings_base = get_project_settings()
        settings_copy = settings_base.copy()
        self.settings = dict(overridden_settings(settings_copy))

        self.path = path
        self.queue_name = queue_name
        self.way = way
        self.async_num = async_num

        self.rabbit_username = None
        self.rabbit_password = None
        self.rabbit_host = None
        self.rabbit_port = None
        self.rabbit_vhost = None

        self.mongo_host = None
        self.mongo_port = None
        self.mongo_database = None
        self.mongo_table = None
        self.mongo_username = None
        self.mongo_password = None
        self.mongo = None

        self.thread = None
        self.count = 0
        self.name = None
        self.channel = None
        self.connector = None
        self.update_machine = None
        # self._settings = settings.__dict__

        self.request = None
        self.loop = None
        self.session = None
        self.loop = None

        self.rabbit = None

        self.RANDOM_USER_AGENT = self.settings.get('RANDOM_USER_AGENT')

        self.RETRY_ENABLED = self.settings.get('RETRY_ENABLED')
        self.RETRY_TIMES = self.settings.get('RETRY_TIMES')

    def get_proxy(self):
        proxy_res = fuclib.getIP()
        if proxy_res:
            return proxy_res['proxy']

    def close(self):
        pass

    def rabbit_connection(self):
        """rabbitmq连接"""
        self.rabbit = RabbitConnect(scrapy_object=self)
        return self.rabbit.connect()

    def produce(self, url=None, params=None, data=None, json=None, charset=None, cookies=None, method='get',
                headers=None, callback="parse", allow_redirects=True, meta=None, not_request=None, priority=1):
        """生产"""
        if not isinstance(callback, str):
            callback = callback.__name__
        if meta is None:
            meta = dict()
        if not_request or (isinstance(url, str) and not search(r"^https?://", url)):
            request = {'not_request': True,
                       'url': url,
                       'callback': callback,
                       'meta': meta
                       }
        else:
            if isinstance(url, dict):
                ret = url
                request = {
                    'url': ret['url'],
                    'params': ret.get('params'),
                    'data': ret.get('data', ret.get('body')),
                    'json': ret.get('json'),
                    'charset': ret.get('charset'),
                    'cookies': ret.get('cookies'),
                    'method': ret.get('method', 'GET'),
                    'headers': ret.get('headers'),
                    'callback': ret.get('callback') if ret.get('callback') else 'parse',
                    'allow_redirects': ret.get('allow_redirects') if ret.get('allow_redirects') else True,
                    'meta': ret.get('meta') if ret.get('meta') else {},
                    'url_encoded': True if '%' in ret['url'] else False,
                    'priority': ret.get('priority', 1),
                    'retry_times': ret.get('retry_times', 0)
                }
            else:
                request = {
                    'url': url,
                    'params': params,
                    'data': data,
                    'json': json,
                    'charset': charset,
                    'cookies': cookies,
                    'method': method,
                    'headers': headers,
                    'callback': callback,
                    'allow_redirects': allow_redirects,
                    'meta': meta,
                    'url_encoded': True if '%' in url else False,
                    'priority': priority
                }
        if self.RANDOM_USER_AGENT:
            request['headers']['User-Agent'] = random_ua()
        if request['data']:
            request['method'] = 'post'
        if self.timeout:
            request['time_out'] = self.timeout
        if self.thread:
            self.rabbit.produce(channel=self.thread.channel, request=dumps(request), queue=self.queue_name,
                                priority=request['priority'])
            pass
        else:
            self.rabbit.produce(request=dumps(request), queue=self.queue_name, priority=request['priority'])
        self.count += 1

        logger.info("(%s) %s 生产：%s" % (self.queue_name, self.count, request))

    def consume(self):
        """消费"""
        while True:
            try:
                # self.connector, self.channel = self.rabbit.sure_conn(self.queue_name, self.connector, self.channel)
                self.rabbit.consume(self.channel, self.queue_name, callback=self.callback,
                                    prefetch_count=self.async_num)

                break
            except exceptions.ConnectionClosed:
                self.connector, self.channel = self.rabbit_connection()
                print('-', "pika.exceptions.ConnectionClosed", ctime())

    async def deal(self, ch, method, properties, ret):
        try:
            self.routing(self.__getattribute__(ret['callback']), ret)
            self.connector.add_callback_threadsafe(partial(ch.basic_ack, method.delivery_tag))
        except Exception as e:
            print_exc()
            self.deal_error(e)

    def routing(self, function, *args, **kwargs):
        result_type = function(*args, **kwargs)
        if isinstance(result_type, Generator):
            for i in result_type:
                if isinstance(i, scrapy.Request):  # 通过yield过来的请求
                    request = self.decode_request(request_to_dict(i, self))
                    self.produce(request)
                elif isinstance(i, scrapy.Item):
                    self.pipeline(i)
                else:
                    raise
        elif isinstance(result_type, scrapy.Request):
            request = self.decode_request(request_to_dict(result_type, self))  # 通过return过来的请求
            self.produce(request)
        elif isinstance(result_type, scrapy.Item) or isinstance(result_type, dict):
            self.pipeline(result_type)

    def pipeline(self, item):
        pipeline_object = getattr(self, "pipelineObj")
        ret = pipeline_object.process_item(item, self)
        logger.debug("(%s) Item %s：%s" % (self.queue_name, pipeline_object.count, item))  # todo

    def decode_request(self, request: dict):
        """从scrapy传过来的Request对象部分键值为byte类型"""
        result = {}
        for k, v in request.items():
            if isinstance(k, bytes):
                new_k = k.decode()
            else:
                new_k = k
            if isinstance(v, bytes):
                new_v = v.decode()
            elif isinstance(v, dict):
                new_v = self.decode_request(v)
            elif isinstance(v, (list, tuple)):
                if len(v) > 1:
                    raise
                if not v:
                    new_v = ''
                else:
                    new_v = v[0].decode() if isinstance(v[0], bytes) else v[0]
            else:
                new_v = v
            result[new_k] = new_v
        return result

    def deal_error(self, e):
        if self.way == 'auto':
            names = ("error_type", "auto_frequency")
            values = (e.args[0], -1)
            logger.error("自动更新报错。", names, values)
        os._exit(1)

    def callback(self, ch, method, properties, body):
        """rabbit_mq回调函数"""
        if body:
            try:
                result = body.decode()
            except UnicodeDecodeError:
                result = body
        else:
            raise
        ret = loads(result)
        ret = self.before_request(ret)
        if 'not_request' in ret:
            logger.info('(%s) 消费：%s' % (self.queue_name, ret))
            coroutine = self.deal(ch, method, properties, ret)
            run_coroutine_threadsafe(coroutine, self.loop)
        else:
            # if not ret['headers']:
            #     ret['headers'] = {"User-Agent": fuclib.ezfuc.random_ua()}  # todo
            if self.proxy:
                proxy_res = fuclib.getIP()
                if proxy_res:
                    ret['proxies'] = proxy_res['proxy']
                    # ret['headers']["Proxy-Authorization"] = fuclib.proxyAuth
            logger.info('(%s) 消费：%s' % (self.queue_name, ret))

            coroutine = self.deal_resp(ch, method, properties, ret)
            run_coroutine_threadsafe(coroutine, self.loop)

    async def deal_resp(self, ch, method, properties, ret):
        """请求并回调处理响应函数"""
        try:
            response = await self.request.quest(self.session, ret)
            if response and (response.status_code in self.allow_status_code or response.status_code == 200):
                try:
                    self.routing(self.__getattribute__(ret['callback']), response)
                except Exception as e:
                    self.connector.add_callback_threadsafe(partial(ch.basic_nack, method.delivery_tag))
                    print_exc()
                    self.deal_error(e)
                self.connector.add_callback_threadsafe(partial(ch.basic_ack, method.delivery_tag))
            else:
                if self.RETRY_ENABLED:  # todo 重试方法需要重写
                    retry_times = ret.get('retry_times', 5)
                    if retry_times <= self.RETRY_TIMES:
                        if response:
                            logger.error("第 %s 次 (%s) 请求失败!返回状态码：%d,返回队列%s" % (
                                retry_times, self.name, response.status_code, ret))
                        else:
                            logger.error(
                                "第 %s 次 (%s) 请求报错!未返回消息,返回队列%s" % (retry_times, self.queue_name, self.queue_name))
                        ret['retry_times'] = retry_times + 1
                        self.produce(ret)
                        self.connector.add_callback_threadsafe(partial(ch.basic_ack, method.delivery_tag))
                else:
                    logger.info("(%s) 请求报错!未返回消息" % self.name)
                    self.produce(ret)
                    self.connector.add_callback_threadsafe(partial(ch.basic_ack, method.delivery_tag))
        except Exception as e:
            print_exc()
            self.deal_error(e)

    def start_requests(self):
        """抽象生产函数"""
        if self.start_urls:
            for url in self.start_urls:
                self.produce(url)

    def parse(self, response):
        """抽象解析函数"""
        pass

    def before_request(self, ret):
        """请求中间件"""
        return ret

    @staticmethod
    def run_forever(loop):
        """实时接收新事件"""
        set_event_loop(loop)
        loop.run_forever()

    def main(self):
        """启动函数"""
        # 配置文件初始化
        serious = [
            'rabbit_username',
            'rabbit_password',
            'rabbit_host',
            'rabbit_port',
            'rabbit_vhost',
            'mongo_host',
            'mongo_port',
            'mongo_database',
            'mongo_table',
            'mongo_username',
            'mongo_password'
        ]
        for i in serious:
            if hasattr(self, i) and getattr(self, i) is None:
                attr = getattr(conf, i)
                setattr(self, i, attr if attr else None)
        self.settings.update(self.custom_settings)
        # self._settings.setdefault('DOWNLOADER_MIDDLEWARES',
        #                           {'DOWNLOADER_MIDDLEWARES': {'frame.tools.middleware.DownloaderMiddleware': 1}})

        self.connector, self.channel = self.rabbit_connection()
        logger.info('rabbitmq配置：%s' % [self.rabbit_host, self.rabbit_port, self.rabbit_username])
        logger.info('mongodb配置：%s' % [self.mongo_host, self.rabbit_port, self.mongo_database])

        # # 脚本信息入库
        # self.update_machine = gethostbyname(gethostname())
        # logger.info("本机ip为：%s" % self.update_machine)
        # todo 保存脚本运行时间

        if self.way and (self.way == 'm' or self.way == 'auto'):
            if self.priority_queue:
                self.channel.queue_declare(queue=self.queue_name, durable=True,
                                           arguments={"x-max-priority": 10})  # todo 优先级队列自动切换
            else:
                self.channel.queue_declare(queue=self.queue_name, durable=True)
            if self.queue_purge:
                self.rabbit.queue_purge(self.queue_name)
            else:
                logger.info("继续生产!!!")
            self.routing(self.start_requests)

        if self.way and (self.way == 'w' or self.way == 'auto'):
            self.request = Request()
            self.loop = new_event_loop()
            self.session = self.loop.run_until_complete(self.request.new_session())

            self.thread = DaemonRun(self.run_forever, self.loop, self)
            self.thread.start()

            self.consume()

            run_coroutine_threadsafe(self.request.exit(self.session), self.loop)

            self.close()

        if not self.way or self.way not in ('m', 'w', 'auto'):
            raise

        self.channel.close()


Spider = Engine



# -*- coding: utf-8 -*-
from requests import get
from pika import PlainCredentials
from pika import BlockingConnection
from pika import ConnectionParameters
from pika import BasicProperties
from logging import getLogger
from scrapy_rabbit.conf import conf

logger = getLogger(__name__)


class RabbitConnect(object):
    """
    RabbitMq消息队列操作
    """
    def __init__(self, scrapy_object=None):
        """
        初始化，并创建连接对象
        """
        self.__rabbit_username = scrapy_object.rabbit_username if scrapy_object.rabbit_username else conf.rabbit_username
        self.__rabbit_password = scrapy_object.rabbit_password if scrapy_object.rabbit_password else conf.rabbit_password
        self.__rabbit_host = scrapy_object.rabbit_host if scrapy_object.rabbit_host else conf.rabbit_host
        self.__rabbit_port = scrapy_object.rabbit_port if scrapy_object.rabbit_port else conf.rabbit_port
        self.__rabbit_vhost = scrapy_object.rabbit_vhost if scrapy_object.rabbit_vhost else conf.rabbit_vhost

        self.priority_queue = scrapy_object.priority_queue if scrapy_object else False

        credentials = PlainCredentials(username=self.__rabbit_username,
                                       password=self.__rabbit_password)
        self.__connector = BlockingConnection(
            ConnectionParameters(host=self.__rabbit_host,
                                 port=self.__rabbit_port,
                                 credentials=credentials,
                                 heartbeat=0),
        )
        self.__channel = self.__connector.channel()

    def connect(self):
        """返回连接对象"""
        return self.__connector, self.__channel

    def produce(self, request, queue, priority, channel=None):
        """生产"""
        if channel:
            channel.basic_publish(
                exchange='',
                routing_key=queue,
                body=request,
                properties=BasicProperties(
                    delivery_mode=2,
                    priority=priority
                )
            )
        else:
            self.__channel.basic_publish(
                exchange='',
                routing_key=queue,
                body=request,
                properties=BasicProperties(
                    delivery_mode=2,
                    priority=priority
                )
            )
            

    def consume(self, channel, queue, callback, prefetch_count, durable=True, max_priority=10):
        """消费"""
        if self.priority_queue:
            channel.queue_declare(queue=queue, durable=durable, arguments={"x-max-priority": max_priority})
        else:
            channel.queue_declare(queue=queue, durable=durable)
        channel.basic_qos(prefetch_count=prefetch_count)
        channel.basic_consume(
            on_message_callback=callback,
            queue=queue,
            auto_ack=False)
        channel.start_consuming()

    def queue_detail(self, queue_name):
        """查看队列当前数据量"""
        queue_name = queue_name.replace('/', '%2F')
        url = f'http://{self.__rabbit_host}:15672/api/queues/{self.__rabbit_vhost}/{queue_name}'
        response = get(url, auth=(self.__rabbit_username, self.__rabbit_password))
        if response.status_code != 200:
            return None
        dic = response.json()
        return {'ready': dic['messages_ready'], 'unconfirmed': dic['messages_unacknowledged'], 'total': dic['messages']}

    def queue_purge(self, queue_name):
        """清空队列"""
        self.__channel.queue_purge(queue_name)
        logger.info(f'{queue_name}  队列已清空')

    def queue_del(self, queue_name, if_unused=False, if_empty=True):
        """删除队列"""
        self.__channel.queue_delete(queue=queue_name, if_unused=if_unused, if_empty=if_empty)
        logger.info(f'{queue_name}  队列已删除')



import os
import shutil
import string

from importlib import import_module
from os.path import join, dirname, abspath, exists, splitext

import scrapy
from scrapy_rabbit.utils.template import render_templatefile
from scrapy_rabbit.commands import ScrapyCommand
from scrapy.utils.template import string_camelcase
from scrapy.exceptions import UsageError


def sanitize_module_name(module_name):
    """Sanitize the given module name, by replacing dashes and points
    with underscores and prefixing it with a letter if it doesn't start
    with one
    """
    module_name = module_name.replace('-', '_').replace('.', '_')
    if module_name[0] not in string.ascii_letters:
        module_name = "a" + module_name
    return module_name


class Command(ScrapyCommand):

    requires_project = False
    default_settings = {'LOG_ENABLED': False}

    def syntax(self):
        return "[options] <class> <name> <domain>"

    def short_desc(self):
        return "Generate new spider using pre-defined templates"

    def add_options(self, parser):
        ScrapyCommand.add_options(self, parser)
        parser.add_option("-l", "--list", dest="list", action="store_true",
            help="List available templates")
        parser.add_option("-e", "--edit", dest="edit", action="store_true",
            help="Edit spider after creating it")
        parser.add_option("-d", "--dump", dest="dump", metavar="TEMPLATE",
            help="Dump template to standard output")
        parser.add_option("-t", "--template", dest="template", default="basic",
            help="Uses a custom template.")
        parser.add_option("--force", dest="force", action="store_true",
            help="If the spider already exists, overwrite it with the template")

    def run(self, args, opts):
        if opts.list:
            self._list_templates()
            return
        if opts.dump:
            template_file = self._find_template(opts.dump)
            if template_file:
                with open(template_file, "r") as f:
                    print(f.read())
            return
        if len(args) != 3:
            raise UsageError()
        class_name, name, domain = args[0:3]
        module = sanitize_module_name(name)

        if self.settings.get('BOT_NAME') == module:
            print("Cannot create a spider with the same name as your project")
            return

        try:
            spidercls = self.crawler_process.spider_loader.load(name)
        except KeyError:
            pass
        else:
            # if spider already exists and not --force then halt
            if not opts.force:
                print("Spider %r already exists in module:" % name)
                print("  %s" % spidercls.__module__)
                return
        template_file = self._find_template(opts.template)
        if template_file:
            self._genspider(class_name, module, name, domain, opts.template, template_file)
            if opts.edit:
                self.exitcode = os.system('scrapy edit "%s"' % name)

    def _genspider(self, class_name, module, name, domain, template_name, template_file):
        """Generate the spider module, based on the given template"""
        tvars = {
            'project_name': self.settings.get('BOT_NAME'),
            'ProjectName': string_camelcase(self.settings.get('BOT_NAME')),
            'module': module,
            'name': name,
            'domain': domain,
            'class_name': class_name,
            'classname': '%sSpider' % ''.join(s.capitalize() for s in module.split('_'))
        }
        if self.settings.get('NEWSPIDER_MODULE'):
            spiders_module = import_module(self.settings['NEWSPIDER_MODULE'])
            spiders_dir = os.path.join(abspath(dirname(spiders_module.__file__)), class_name)
            if not os.path.exists(spiders_dir):
                os.makedirs(spiders_dir)
        else:
            spiders_module = None
            spiders_dir = "."

        spider_file = "%s.py" % join(spiders_dir, module)
        shutil.copyfile(template_file, spider_file)
        render_templatefile(spider_file, **tvars)
        print("Created spider %r using template %r "
              % (name, template_name), end=('' if spiders_module else '\n'))
        if spiders_module:
            print("in module:\n  %s.%s" % (spiders_module.__name__, module))

    def _find_template(self, template):
        template_file = join(self.templates_dir, '%s.tmpl' % template)
        if exists(template_file):
            return template_file
        print("Unable to find template: %s\n" % template)
        print('Use "scrapy genspider --list" to see all available templates.')

    def _list_templates(self):
        print("Available templates:")
        for filename in sorted(os.listdir(self.templates_dir)):
            if filename.endswith('.tmpl'):
                print("  %s" % splitext(filename)[0])

    @property
    def templates_dir(self):
        _templates_base_dir = self.settings['TEMPLATES_DIR'] or \
            join(scrapy.__path__[0], 'templates')
        return join(_templates_base_dir, 'spiders')



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
import aiohttp
import chardet
import asyncio
import fuclib
import parsel
from urllib.parse import urljoin
from yarl import URL
from w3lib.encoding import (html_to_unicode, http_content_type_encoding)
import json as js

import logging
from logging import error

logger = logging.getLogger(__name__)


def random_ua():
    """
    随机浏览器头
    :return:
    """
    import random
    user_agent_list = [
        'Mozilla/5.0 (Windows NT {WindowsNT};{WOW64}{language} rv:{Firefox}) Gecko/{builddata} Firefox/{Firefox}'.format(
            **{'WindowsNT': random.choice(["6.1", "6.2", "6.3", "10.0"]),
               'WOW64': random.choice(["", " WOW64;", " Win64;", " x64;"]),
               'language': random.choice(["", " {};".format("zh-CN")]), 'builddata': random.choice(
                    ["201{}0{}{}".format(random.randint(0, 6), random.randint(1, 9), random.randint(10, 28))]),
               'Firefox': random.choice(
                   ["50.0.1", "50.0.2", "50.0", "50.01", "50.010", "50.011", "50.02", "50.03", "50.04", "50.05",
                    "50.06", "50.07", "50.08", "50.09", "50.1.0", "51.0.1", "51.0", "51.01", "51.010", "51.011",
                    "51.012", "51.013", "51.014", "51.02", "51.03", "51.04", "51.05", "51.06", "51.07", "51.08",
                    "51.09", "52.0.1", "52.0.2", "52.0", "52.01", "52.02", "52.03", "52.04", "52.05", "52.06",
                    "52.07", "52.08", "52.09", "52.1.0", "52.1.1", "52.1.2", "52.2.0", "52.2.1", "52.3.0", "52.4.0",
                    "52.4.1", "53.0.2", "53.0.3", "53.0", "53.01", "53.010", "53.02", "53.03", "53.04", "53.05",
                    "53.06", "53.07", "53.08", "53.09", "54.0.1", "54.0", "54.01", "54.010", "54.011", "54.012",
                    "54.013", "54.02", "54.03", "54.04", "54.05", "54.06", "54.07", "54.08", "54.09", "55.0.1",
                    "55.0.2", "55.0.3", "55.0", "55.01", "55.010", "55.011", "55.012", "55.013", "55.02", "55.03",
                    "55.04", "55.05", "55.06", "55.07", "55.08", "55.09", "56.0.1", "56.0", "56.01", "56.010",
                    "56.011", "56.012", "56.02", "56.03", "56.04", "56.05", "56.06", "56.07", "56.08", "56.09",
                    "57.03", "57.04", "57.05", "57.06"]), }),
        'Mozilla/5.0 (Windows NT {WindowsNT};{WOW64}{language}) AppleWebKit/{Safari} (KHTML, '
        'like Gecko) Chrome/{Chrome} Safari/{Safari}'.format(
            **{'WindowsNT': random.choice(["6.1", "6.2", "6.3", "10"]),
               'WOW64': random.choice(["", " WOW64;", " Win64;", " x64;"]),
               'language': random.choice(["", " {};".format("zh-CN")]),
               'Chrome': '{0}.{1}.{2}.{3}'.format(random.randint(50, 61), random.randint(0, 9),
                                                  random.randint(1000, 9999), random.randint(10, 99)),
               'Safari': '{0}.{1}'.format(random.randint(100, 999), random.randint(0, 99)), }),
        'Mozilla/5.0 ({compatible}Windows NT {WindowsNT};{WOW64} MSIE {ie}.0; Trident/{Trident}.0;){Gecko}'.format(
            **{'compatible': random.choice(["", "compatible; "]),
               'WindowsNT': random.choice(["6.1", "6.2", "6.3", "10"]),
               'WOW64': random.choice(["", " WOW64;", " Win64;", " x64;"]),
               'ie': random.randint(10, 11), 'Trident': random.randint(5, 7),
               'Gecko': random.choice(["", " like Gecko;"])}),
        'Mozilla/5.0 (Windows NT {WindowsNT}; MSIE 9.0;) Opera {opera1}.{opera2}'.format(
            **{'WindowsNT': random.choice(["6.1", "6.2", "6.3", "10"]),
               'opera1': random.randint(10, 12), 'opera2': random.randint(10, 99)}),
    ]
    rs = random.choice(user_agent_list)  # 201706  firefox14 chrome63 ie9 opera2
    return rs


class Response:
    def __init__(self, data, content=None, status_code=None, charset=None, cookies=None, method=None,
                 headers=None, callback="parse", proxies=None, meta=None, res=None):
        if meta is None:
            meta = {}
        self.url = str(res.url)
        self.data = data
        self.body = self.content = content
        self.status = self.status_code = status_code
        self.charset = charset
        self.res = res
        self.cookies = {k: v.value for k, v in self.res.cookies.items()}
        self.method = method
        self.headers = headers
        self.callback = callback
        self.proxies = proxies
        self.meta = meta if meta else {}
        self.__r = parsel.Selector(self.text)

    def __str__(self):
        return self.text

    @property
    def text(self):
        if not self.content:
            return ''
        if self.charset:
            try:
                text = self.content.decode(self.charset)
            except UnicodeDecodeError:
                try:
                    benc = http_content_type_encoding(dict(self.res.headers)['Content-Type'])
                    if benc:
                        charset = 'charset=%s' % benc
                        text = html_to_unicode(charset, self.body)[1]
                    else:
                        raise UnicodeDecodeError
                except UnicodeDecodeError:
                    try:
                        char = chardet.detect(self.content)
                        if char:
                            text = self.content.decode(char['encoding'])
                        else:
                            raise UnicodeDecodeError
                    except UnicodeDecodeError:
                        try:
                            text = self.content.decode('utf-8')
                        except UnicodeDecodeError:
                            try:
                                text = self.content.decode("gb18030")
                            except UnicodeDecodeError:
                                text = self.content.decode('utf-8', "ignore")
        else:
            try:
                text = self.content.decode('utf-8')
            except UnicodeDecodeError:
                try:
                    char = chardet.detect(self.content)
                    if char:
                        text = self.content.decode(char['encoding'])
                    else:
                        raise UnicodeDecodeError
                except UnicodeDecodeError:
                    try:
                        text = self.content.decode('gb18030')
                    except UnicodeDecodeError:
                        text = self.content.decode('utf-8', "ignore")
        return text

    @property
    def json(self):
        return js.loads(self.text, strict=False)

    def xpath(self, x):
        return self.__r.xpath(x)

    def css(self, x):
        return self.__r.css(x)

    def urljoin(self, url):
        return urljoin(self.url, url)


class Retry(object):
    def __init__(self, max_times=3):
        self.func = None
        self.max_times = max_times

    async def run(self, session, request):
        retry_times = 0
        while retry_times < self.max_times:
            try:
                return await self.func(session, request)
            except TimeoutError:
                error(f"第{retry_times + 1}次请求超时，url：{request['url']}")
                retry_times += 1
            except Exception as e:
                error(f"第 {retry_times + 1} 次请求报错({type(e).__name__}: {e})，url：{request['url']}")
                retry_times += 1

    def __call__(self, func):
        self.func = func
        return self.run


class Request:
    @staticmethod
    @Retry()
    async def quest(session, request):
        if request['method'].upper() == 'GET':
            async with session.get(URL(request['url'], encoded=request.get('url_encoded', False)),
                                   params=request.get('params', None), cookies=request.get('cookies', None),
                                   headers=request.get('headers', None), proxy=request.get('proxies', None),
                                   allow_redirects=request.get('allow_redirects', True),
                                   timeout=request.get('time_out', 180),
                                   proxy_auth=aiohttp.BasicAuth(fuclib.proxyUser, fuclib.proxyPass)) as res:
                text = await res.read()

        elif request['method'].upper() == 'POST':
            async with session.post(URL(request['url'], encoded=request.get('url_encoded', False)),
                                    data=request.get('data', None), json=request.get('json', None),
                                    cookies=request.get('cookies', None), headers=request.get('headers', None),
                                    proxy=request.get('proxies', None),
                                    allow_redirects=request.get('allow_redirects', True),
                                    timeout=request.get('time_out', 180),
                                    proxy_auth=aiohttp.BasicAuth(fuclib.proxyUser, fuclib.proxyPass)) as res:
                text = await res.read()
        else:
            raise "{%s}请求方式未定义，请自定义添加！" % request['method']

        if res:
            status_code = res.status
            charset = res.charset
            response = Response(request.get('data'), text, status_code, charset, request.get('cookies', None),
                                request['method'], request.get('headers', None), request.get('callback', "parse"),
                                request.get('proxies', None), request.get('meta', None), res)
            return response

    @staticmethod
    async def new_session(verify=False):
        connector = aiohttp.TCPConnector(ssl=verify)
        session = aiohttp.ClientSession(connector=connector, trust_env=True)
        return session

    @staticmethod
    async def get(session, request):
        async with session.get(request['url'], params=request['params'], cookies=request['cookies'],
                               headers=request['headers'], proxy=request['proxies'],
                               allow_redirects=request['allow_redirects'], timeout=request.get('time_out', 180)) as res:
            text = await res.read()
            return res, text

    @staticmethod
    async def post(session, request):
        async with session.post(request['url'], data=request['data'], json=request['json'], cookies=request['cookies'],
                                headers=request['headers'], proxy=request['proxies'],
                                allow_redirects=request['allow_redirects']) as res:
            text = await res.read()
            return res, text

    @staticmethod
    async def exit(session):
        await session.close()

    @staticmethod
    def func(future):
        logger.info(future.result())
        return future.result()



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
"""
Base class for Scrapy commands
"""
import os
from optparse import OptionGroup
from twisted.python import failure

from scrapy.utils.conf import arglist_to_dict
from scrapy.exceptions import UsageError


class ScrapyCommand:
    requires_project = False
    crawler_process = None

    # default settings to be used for this command instead of global defaults
    default_settings = {}

    exitcode = 0

    def __init__(self):
        self.settings = None  # set in scrapy.cmdline

    def set_crawler(self, crawler):
        if hasattr(self, '_crawler'):
            raise RuntimeError("crawler already set")
        self._crawler = crawler

    def syntax(self):
        """
        Command syntax (preferably one-line). Do not include command name.
        """
        return ""

    def short_desc(self):
        """
        A short description of the command
        """
        return ""

    def long_desc(self):
        """A long description of the command. Return short description when not
        available. It cannot contain newlines, since contents will be formatted
        by optparser which removes newlines and wraps text.
        """
        return self.short_desc()

    def help(self):
        """An extensive help for the command. It will be shown when using the
        "help" command. It can contain newlines, since no post-formatting will
        be applied to its contents.
        """
        return self.long_desc()

    def add_options(self, parser):
        """
        Populate option parse with options available for this command
        """
        group = OptionGroup(parser, "Global Options")
        group.add_option("--logfile", metavar="FILE",
                         help="log file. if omitted stderr will be used")
        group.add_option("-L", "--loglevel", metavar="LEVEL", default=None,
                         help="log level (default: %s)" % self.settings['LOG_LEVEL'])
        group.add_option("--nolog", action="store_true",
                         help="disable logging completely")
        group.add_option("--profile", metavar="FILE", default=None,
                         help="write python cProfile stats to FILE")
        group.add_option("--pidfile", metavar="FILE",
                         help="write process ID to FILE")
        group.add_option("-s", "--set", action="append", default=[], metavar="NAME=VALUE",
                         help="set/override setting (may be repeated)")
        group.add_option("--pdb", action="store_true", help="enable pdb on failure")

        parser.add_option_group(group)

    def process_options(self, args, opts):
        try:
            self.settings.setdict(arglist_to_dict(opts.set),
                                  priority='cmdline')
        except ValueError:
            raise UsageError("Invalid -s value, use -s NAME=VALUE", print_help=False)

        if opts.logfile:
            self.settings.set('LOG_ENABLED', True, priority='cmdline')
            self.settings.set('LOG_FILE', opts.logfile, priority='cmdline')

        if opts.loglevel:
            self.settings.set('LOG_ENABLED', True, priority='cmdline')
            self.settings.set('LOG_LEVEL', opts.loglevel, priority='cmdline')

        if opts.nolog:
            self.settings.set('LOG_ENABLED', False, priority='cmdline')

        if opts.pidfile:
            with open(opts.pidfile, "w") as f:
                f.write(str(os.getpid()) + os.linesep)

        if opts.pdb:
            failure.startDebugMode()

    def run(self, args, opts):
        """
        Entry point for running commands
        """
        raise NotImplementedError
